// This source file is generated automatically from dri2.xml

use crate::base::{self, BaseError, BaseEvent, GeEvent, Raw, Reply, WiredIn, WiredOut, Xid};
use crate::ext;
use crate::ffi::base::*;
use crate::ffi::ext::*;
use crate::lat1_str::{Lat1Str, Lat1String, Lat1StrF};
use crate::xproto;
use crate::xproto::PropEl;

use bitflags::bitflags;
use libc::{self, iovec};
use std::convert::TryInto;
use std::hash::{Hash, Hasher};
use std::os::unix::io::RawFd;

/// The official identifier for the `DRI2` extension.
pub const XNAME: &str = "DRI2";
/// The major version of the `DRI2` extension.
pub const MAJOR_VERSION: u32 = 1;
/// The minor version of the `DRI2` extension.
pub const MINOR_VERSION: u32 = 4;
/// The version string of the `DRI2` extension.
pub const VERSION_STRING: &str = "1.4";

pub(crate) static mut FFI_EXT: xcb_extension_t = xcb_extension_t {
    name: "DRI2\0".as_ptr() as *const _,
    global_id: 0,
};

/// Prefetch server runtime info data of the `DRI2` extension.
pub fn prefetch_extension_data(conn: &base::Connection) {
    unsafe {
        xcb_prefetch_extension_data(conn.get_raw_conn(), std::ptr::addr_of_mut!(FFI_EXT));
    }
}

/// Fetch server runtime info data of the `DRI2` extension.
///
/// Might be non-blocking if [prefetch_extension_data] was called before.
/// This function is of seldom use as the extensions are initialized by the
/// [Connection](crate::Connection) constructor.
pub fn get_extension_data(conn: &base::Connection) -> std::option::Option<ext::ExtensionData> {
    unsafe {
        let reply = xcb_get_extension_data(conn.get_raw_conn(), std::ptr::addr_of_mut!(FFI_EXT));
        assert!(!reply.is_null(), "Could not fetch DRI2 extension data");
        let reply = xproto::QueryExtensionReply::from_raw(reply);
        if !reply.present() {
            std::mem::forget(reply);
            return None;
        }
        let res = ext::ExtensionData{
            ext: ext::Extension::Dri2,
            major_opcode: reply.major_opcode(),
            first_event: reply.first_event(),
            first_error: reply.first_error(),
        };
        std::mem::forget(reply);
        Some(res)
    }
}

/// The `BufferSwapCompleteEvent` event.
pub struct BufferSwapCompleteEvent {
    raw: *mut xcb_generic_event_t,
}

impl base::Raw<xcb_generic_event_t> for BufferSwapCompleteEvent {
    unsafe fn from_raw(raw: *mut xcb_generic_event_t) -> Self { BufferSwapCompleteEvent { raw } }

    fn as_raw(&self) -> *mut xcb_generic_event_t {
        self.raw
    }
}

impl base::BaseEvent for BufferSwapCompleteEvent {
    const EXTENSION: std::option::Option<ext::Extension> = Some(ext::Extension::Dri2);
    const NUMBER: u32 = 0;
}

impl BufferSwapCompleteEvent {
    pub fn new(event_base: u8,
        event_type: EventType,
        drawable: xproto::Drawable,
        ust_hi: u32,
        ust_lo: u32,
        msc_hi: u32,
        msc_lo: u32,
        sbc: u32,
    ) -> BufferSwapCompleteEvent {
        unsafe {
            let ptr = libc::malloc(32) as *mut u8;
            let wire_buf = std::slice::from_raw_parts_mut(ptr, 32);
            let mut wire_off = 0usize;
            let response_type = event_base;
            let sequence = 0u16;

            wire_off += response_type.serialize(&mut wire_buf[wire_off ..]);
            wire_off += 1usize;
            wire_off += sequence.serialize(&mut wire_buf[wire_off ..]);
            wire_off += (std::mem::transmute::<_, u32>(event_type) as u16).serialize(&mut wire_buf[wire_off ..]);
            wire_off += 2usize;
            wire_off += drawable.serialize(&mut wire_buf[wire_off ..]);
            wire_off += ust_hi.serialize(&mut wire_buf[wire_off ..]);
            wire_off += ust_lo.serialize(&mut wire_buf[wire_off ..]);
            wire_off += msc_hi.serialize(&mut wire_buf[wire_off ..]);
            wire_off += msc_lo.serialize(&mut wire_buf[wire_off ..]);
            sbc.serialize(&mut wire_buf[wire_off ..]);

            BufferSwapCompleteEvent::from_raw(ptr as *mut xcb_generic_event_t)
        }
    }

    fn wire_ptr(&self) -> *const u8 { self.raw as *const u8 }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn event_type(&self) -> EventType {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            let val = base::value_from_ptr(ptr) as u32;
            std::mem::transmute::<u32, EventType>(val)
        }
    }


    fn drawable(&self) -> xproto::Drawable {
        unsafe {
            let offset = 8usize;
            let res_id = *(self.wire_ptr().add(offset) as *const u32);
            xproto::Drawable::Unknown(res_id)
        }
    }

    pub fn ust_hi(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_lo(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_hi(&self) -> u32 {
        unsafe {
            let offset = 20usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_lo(&self) -> u32 {
        unsafe {
            let offset = 24usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc(&self) -> u32 {
        unsafe {
            let offset = 28usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl std::fmt::Debug for BufferSwapCompleteEvent {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("BufferSwapCompleteEvent")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("event_type", &self.event_type())
            .field("pad", &2)
            .field("drawable", &self.drawable())
            .field("ust_hi", &self.ust_hi())
            .field("ust_lo", &self.ust_lo())
            .field("msc_hi", &self.msc_hi())
            .field("msc_lo", &self.msc_lo())
            .field("sbc", &self.sbc())
            .finish()
    }
}

impl base::WiredOut for BufferSwapCompleteEvent {
    fn wire_len(&self) -> usize {
        32usize
    }

    fn serialize(&self, wire_buf: &mut[u8]) -> usize {
        debug_assert!(wire_buf.len() >= self.wire_len());
        let raw_slice = unsafe { std::slice::from_raw_parts(self.raw as *const u8, self.wire_len()) };
        wire_buf[0 .. self.wire_len()].copy_from_slice(raw_slice);
        self.wire_len()
    }
}

impl base::WiredIn for BufferSwapCompleteEvent {
    type Params = ();

    unsafe fn compute_wire_len(_ptr: *const u8, _params: ()) -> usize {
        32
    }

    unsafe fn unserialize(ptr: *const u8, _params: (), offset: &mut usize) -> Self {
        let sz = Self::compute_wire_len(ptr, ());
        *offset += sz;
        let raw = libc::malloc(sz) as *mut xcb_generic_event_t;
        std::ptr::copy(ptr as *const xcb_generic_event_t, raw, sz);
        BufferSwapCompleteEvent { raw }
    }
}

impl Drop for BufferSwapCompleteEvent {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl Send for BufferSwapCompleteEvent {}
unsafe impl Sync for BufferSwapCompleteEvent {}

/// The `InvalidateBuffersEvent` event.
pub struct InvalidateBuffersEvent {
    raw: *mut xcb_generic_event_t,
}

impl base::Raw<xcb_generic_event_t> for InvalidateBuffersEvent {
    unsafe fn from_raw(raw: *mut xcb_generic_event_t) -> Self { InvalidateBuffersEvent { raw } }

    fn as_raw(&self) -> *mut xcb_generic_event_t {
        self.raw
    }
}

impl base::BaseEvent for InvalidateBuffersEvent {
    const EXTENSION: std::option::Option<ext::Extension> = Some(ext::Extension::Dri2);
    const NUMBER: u32 = 1;
}

impl InvalidateBuffersEvent {
    pub fn new(event_base: u8,
        drawable: xproto::Drawable,
    ) -> InvalidateBuffersEvent {
        unsafe {
            let ptr = libc::malloc(32) as *mut u8;
            let wire_buf = std::slice::from_raw_parts_mut(ptr, 32);
            let mut wire_off = 0usize;
            let response_type = 1u8 + event_base;
            let sequence = 0u16;

            wire_off += response_type.serialize(&mut wire_buf[wire_off ..]);
            wire_off += 1usize;
            wire_off += sequence.serialize(&mut wire_buf[wire_off ..]);
            drawable.serialize(&mut wire_buf[wire_off ..]);

            InvalidateBuffersEvent::from_raw(ptr as *mut xcb_generic_event_t)
        }
    }

    fn wire_ptr(&self) -> *const u8 { self.raw as *const u8 }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    fn drawable(&self) -> xproto::Drawable {
        unsafe {
            let offset = 4usize;
            let res_id = *(self.wire_ptr().add(offset) as *const u32);
            xproto::Drawable::Unknown(res_id)
        }
    }
}

impl std::fmt::Debug for InvalidateBuffersEvent {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("InvalidateBuffersEvent")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("drawable", &self.drawable())
            .finish()
    }
}

impl base::WiredOut for InvalidateBuffersEvent {
    fn wire_len(&self) -> usize {
        32usize
    }

    fn serialize(&self, wire_buf: &mut[u8]) -> usize {
        debug_assert!(wire_buf.len() >= self.wire_len());
        let raw_slice = unsafe { std::slice::from_raw_parts(self.raw as *const u8, self.wire_len()) };
        wire_buf[0 .. self.wire_len()].copy_from_slice(raw_slice);
        self.wire_len()
    }
}

impl base::WiredIn for InvalidateBuffersEvent {
    type Params = ();

    unsafe fn compute_wire_len(_ptr: *const u8, _params: ()) -> usize {
        32
    }

    unsafe fn unserialize(ptr: *const u8, _params: (), offset: &mut usize) -> Self {
        let sz = Self::compute_wire_len(ptr, ());
        *offset += sz;
        let raw = libc::malloc(sz) as *mut xcb_generic_event_t;
        std::ptr::copy(ptr as *const xcb_generic_event_t, raw, sz);
        InvalidateBuffersEvent { raw }
    }
}

impl Drop for InvalidateBuffersEvent {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl Send for InvalidateBuffersEvent {}
unsafe impl Sync for InvalidateBuffersEvent {}

/// Unified event type for the Dri2 extension
#[derive(Debug)]
pub enum Event {
    BufferSwapComplete(BufferSwapCompleteEvent),
    InvalidateBuffers(InvalidateBuffersEvent),
}

impl Event {
  pub fn as_raw(&self) -> *mut xcb_generic_event_t {
    match self {
      Self::BufferSwapComplete(e) => e.as_raw(),
      Self::InvalidateBuffers(e) => e.as_raw(),
    }
  }
}

impl base::ResolveWireEvent for Event {
    unsafe fn resolve_wire_event(first_event: u8, raw: *mut xcb_generic_event_t) -> std::option::Option<Self> {
        debug_assert!(!raw.is_null());
        let response_type = (*raw).response_type & 0x7F;
        debug_assert!(response_type != 0, "This is not an event but an error!");
        debug_assert!(response_type != XCB_GE_GENERIC, "This is a GE_GENERIC event!");
        match response_type - first_event {
            0 => std::option::Option::Some(Event::BufferSwapComplete(BufferSwapCompleteEvent::from_raw(raw))),
            1 => std::option::Option::Some(Event::InvalidateBuffers(InvalidateBuffersEvent::from_raw(raw))),
            _ => std::option::Option::None,
        }
    }
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
#[repr(u32)]
pub enum Attachment {
    BufferFrontLeft = 0,
    BufferBackLeft = 1,
    BufferFrontRight = 2,
    BufferBackRight = 3,
    BufferDepth = 4,
    BufferStencil = 5,
    BufferAccum = 6,
    BufferFakeFrontLeft = 7,
    BufferFakeFrontRight = 8,
    BufferDepthStencil = 9,
    BufferHiz = 10,
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
#[repr(u32)]
pub enum DriverType {
    Dri = 0,
    Vdpau = 1,
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
#[repr(u32)]
pub enum EventType {
    ExchangeComplete = 1,
    BlitComplete = 2,
    FlipComplete = 3,
}

#[derive(Copy, Clone)]
pub struct Dri2Buffer {
    data: [u8; 20],
}

#[allow(unused_parens)]
impl Dri2Buffer {
    pub(crate) unsafe fn from_data<D: AsRef<[u8]> + ?Sized>(data: &D) -> &Dri2Buffer {
        debug_assert_eq!(data.as_ref().len(), 20);
        &*(data.as_ref() as *const [u8] as *const Dri2Buffer)
    }

    /// Construct a new [Dri2Buffer].
    #[allow(unused_assignments, unused_unsafe)]
    pub fn new(
        attachment: Attachment,
        name: u32,
        pitch: u32,
        cpp: u32,
        flags: u32,
    ) -> Dri2Buffer {
            unsafe {
            let mut wire_buf = [0u8; 20];
            let mut wire_off = 0usize;

            wire_off += std::mem::transmute::<_, u32>(attachment).serialize(&mut wire_buf[wire_off .. ]);
            wire_off += name.serialize(&mut wire_buf[wire_off .. ]);
            wire_off += pitch.serialize(&mut wire_buf[wire_off .. ]);
            wire_off += cpp.serialize(&mut wire_buf[wire_off .. ]);
            wire_off += flags.serialize(&mut wire_buf[wire_off .. ]);

            Dri2Buffer { data: wire_buf }
        }
    }

    fn wire_ptr(&self) -> *const u8 { self.data.as_ptr() }

    fn wire_len(&self) -> usize { self.data.len() }

    pub fn attachment(&self) -> Attachment {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            let val = base::value_from_ptr(ptr) as u32;
            std::mem::transmute::<u32, Attachment>(val)
        }
    }

    pub fn name(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn pitch(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn cpp(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn flags(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

#[test]
fn test_sizeof_dri2_buffer() {
    assert_eq!(std::mem::size_of::<Dri2Buffer>(), 20);
}

impl base::WiredOut for Dri2Buffer {
    fn wire_len(&self) -> usize { 20 }

    fn serialize(&self, wire_buf: &mut [u8]) -> usize {
        (&mut wire_buf[..self.data.len()]).copy_from_slice(&self.data);
        self.data.len()
    }
}

impl base::WiredIn for Dri2Buffer {
    type Params = ();
    unsafe fn compute_wire_len(_ptr: *const u8, _params: ()) -> usize { 20 }
    unsafe fn unserialize(ptr: *const u8, _params: (), offset: &mut usize) -> Self {
        *offset += 20;
        *(ptr as *const Dri2Buffer)
    }
}

impl std::fmt::Debug for Dri2Buffer {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Dri2Buffer")
            .field("attachment", &self.attachment())
            .field("name", &self.name())
            .field("pitch", &self.pitch())
            .field("cpp", &self.cpp())
            .field("flags", &self.flags())
            .finish()
    }
}

#[derive(Copy, Clone)]
pub struct AttachFormat {
    data: [u8; 8],
}

#[allow(unused_parens)]
impl AttachFormat {
    pub(crate) unsafe fn from_data<D: AsRef<[u8]> + ?Sized>(data: &D) -> &AttachFormat {
        debug_assert_eq!(data.as_ref().len(), 8);
        &*(data.as_ref() as *const [u8] as *const AttachFormat)
    }

    /// Construct a new [AttachFormat].
    #[allow(unused_assignments, unused_unsafe)]
    pub fn new(
        attachment: Attachment,
        format: u32,
    ) -> AttachFormat {
            unsafe {
            let mut wire_buf = [0u8; 8];
            let mut wire_off = 0usize;

            wire_off += std::mem::transmute::<_, u32>(attachment).serialize(&mut wire_buf[wire_off .. ]);
            wire_off += format.serialize(&mut wire_buf[wire_off .. ]);

            AttachFormat { data: wire_buf }
        }
    }

    fn wire_ptr(&self) -> *const u8 { self.data.as_ptr() }

    fn wire_len(&self) -> usize { self.data.len() }

    pub fn attachment(&self) -> Attachment {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            let val = base::value_from_ptr(ptr) as u32;
            std::mem::transmute::<u32, Attachment>(val)
        }
    }

    pub fn format(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

#[test]
fn test_sizeof_attach_format() {
    assert_eq!(std::mem::size_of::<AttachFormat>(), 8);
}

impl base::WiredOut for AttachFormat {
    fn wire_len(&self) -> usize { 8 }

    fn serialize(&self, wire_buf: &mut [u8]) -> usize {
        (&mut wire_buf[..self.data.len()]).copy_from_slice(&self.data);
        self.data.len()
    }
}

impl base::WiredIn for AttachFormat {
    type Params = ();
    unsafe fn compute_wire_len(_ptr: *const u8, _params: ()) -> usize { 8 }
    unsafe fn unserialize(ptr: *const u8, _params: (), offset: &mut usize) -> Self {
        *offset += 8;
        *(ptr as *const AttachFormat)
    }
}

impl std::fmt::Debug for AttachFormat {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("AttachFormat")
            .field("attachment", &self.attachment())
            .field("format", &self.format())
            .finish()
    }
}

pub(crate) fn request_name(opcode: u16) -> std::option::Option<&'static str> {
    match opcode {
        0 => Some("dri2::QueryVersion"),
        1 => Some("dri2::Connect"),
        2 => Some("dri2::Authenticate"),
        3 => Some("dri2::CreateDrawable"),
        4 => Some("dri2::DestroyDrawable"),
        5 => Some("dri2::GetBuffers"),
        6 => Some("dri2::CopyRegion"),
        7 => Some("dri2::GetBuffersWithFormat"),
        8 => Some("dri2::SwapBuffers"),
        9 => Some("dri2::GetMsc"),
        10 => Some("dri2::WaitMsc"),
        11 => Some("dri2::WaitSbc"),
        12 => Some("dri2::SwapInterval"),
        13 => Some("dri2::GetParam"),
        _ => None,
    }
}

/// Reply type for [QueryVersion].
///
/// Can be obtained from a [QueryVersionCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [QueryVersionCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct QueryVersionReply {
    raw: *const u8,
}

impl QueryVersionReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn major_version(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn minor_version(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for QueryVersionReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for QueryVersionReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("QueryVersionReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("major_version", &self.major_version())
            .field("minor_version", &self.minor_version())
            .finish()
    }
}

impl Drop for QueryVersionReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for QueryVersionReply {}
unsafe impl std::marker::Sync for QueryVersionReply {}

/// Cookie type for [QueryVersion].
///
/// This cookie can be used to get a [QueryVersionReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct QueryVersionCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [QueryVersion].
///
/// This cookie can be used to get a [QueryVersionReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct QueryVersionCookieUnchecked {
    seq: u64,
}

impl base::Cookie for QueryVersionCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        QueryVersionCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for QueryVersionCookie {
}

unsafe impl base::CookieWithReplyChecked for QueryVersionCookie {
    type Reply = QueryVersionReply;
}

impl base::Cookie for QueryVersionCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        QueryVersionCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for QueryVersionCookieUnchecked {
    type Reply = QueryVersionReply;
}

/// The `QueryVersion` request.
///
/// This request replies [QueryVersionReply].
///
/// Associated cookie types are [QueryVersionCookie] and [QueryVersionCookieUnchecked].
#[derive(Clone, Debug)]
pub struct QueryVersion {
    pub major_version: u32,
    pub minor_version: u32,
}

unsafe impl base::RawRequest for QueryVersion {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 0,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 12];
        self.major_version.serialize(&mut buf0[4 .. ]);
        self.minor_version.serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for QueryVersion {
    type Cookie = QueryVersionCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for QueryVersion {
    type Reply = QueryVersionReply;
    type Cookie = QueryVersionCookie;
    type CookieUnchecked = QueryVersionCookieUnchecked;
}

/// Reply type for [Connect].
///
/// Can be obtained from a [ConnectCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [ConnectCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct ConnectReply {
    raw: *const u8,
}

impl ConnectReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    unsafe fn compute_len(ptr: *const u8, _params: ()) -> usize {
        let mut sz = 0;
        // response_type
        sz += 1usize;
        // pad
        sz += 1usize;
        // sequence
        sz += 2usize;
        // length
        sz += 4usize;
        // driver_name_length
        let driver_name_length = *(ptr.add(sz) as *const u32);
        sz += 4usize;
        // device_name_length
        let device_name_length = *(ptr.add(sz) as *const u32);
        sz += 4usize;
        // pad
        sz += 16usize;
        // driver_name
        sz += (driver_name_length as usize);
        // alignment_pad
        sz += ((((driver_name_length as usize) + 3usize) & (!3usize)) - (driver_name_length as usize));
        // device_name
        sz += (device_name_length as usize);
        sz
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    fn driver_name_length(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    fn device_name_length(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }



    pub fn driver_name(&self) -> &Lat1Str {
        unsafe {
            let offset = 32usize;
            let len = (self.driver_name_length() as usize) as _;
            let ptr = self.wire_ptr().add(offset);
            let bytes = std::slice::from_raw_parts(ptr, len);
            Lat1Str::from_bytes(bytes)
        }
    }

    pub fn alignment_pad(&self) -> &[u8] {
        unsafe {
            let offset = (32usize + (self.driver_name_length() as usize));
            let len = ((((self.driver_name_length() as usize) + 3usize) & (!3usize)) - (self.driver_name_length() as usize));
            let ptr = self.wire_ptr().add(offset) as *const u8;
            std::slice::from_raw_parts(ptr, len)
        }
    }


    pub fn device_name(&self) -> &Lat1Str {
        unsafe {
            let offset = ((32usize + (self.driver_name_length() as usize)) + ((((self.driver_name_length() as usize) + 3usize) & (!3usize)) - (self.driver_name_length() as usize)));
            let len = (self.device_name_length() as usize) as _;
            let ptr = self.wire_ptr().add(offset);
            let bytes = std::slice::from_raw_parts(ptr, len);
            Lat1Str::from_bytes(bytes)
        }
    }
}

impl base::Reply for ConnectReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for ConnectReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("ConnectReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("driver_name_length", &self.driver_name_length())
            .field("device_name_length", &self.device_name_length())
            .field("pad", &16)
            .field("driver_name", &self.driver_name())
            .field("alignment_pad", &self.alignment_pad())
            .field("device_name", &self.device_name())
            .finish()
    }
}

impl Drop for ConnectReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for ConnectReply {}
unsafe impl std::marker::Sync for ConnectReply {}

/// Cookie type for [Connect].
///
/// This cookie can be used to get a [ConnectReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct ConnectCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [Connect].
///
/// This cookie can be used to get a [ConnectReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct ConnectCookieUnchecked {
    seq: u64,
}

impl base::Cookie for ConnectCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        ConnectCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for ConnectCookie {
}

unsafe impl base::CookieWithReplyChecked for ConnectCookie {
    type Reply = ConnectReply;
}

impl base::Cookie for ConnectCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        ConnectCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for ConnectCookieUnchecked {
    type Reply = ConnectReply;
}

/// The `Connect` request.
///
/// This request replies [ConnectReply].
///
/// Associated cookie types are [ConnectCookie] and [ConnectCookieUnchecked].
#[derive(Clone, Debug)]
pub struct Connect {
    pub window: xproto::Window,
    pub driver_type: DriverType,
}

unsafe impl base::RawRequest for Connect {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 1,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 12];
        self.window.serialize(&mut buf0[4 .. ]);
        std::mem::transmute::<_, u32>(self.driver_type).serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for Connect {
    type Cookie = ConnectCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for Connect {
    type Reply = ConnectReply;
    type Cookie = ConnectCookie;
    type CookieUnchecked = ConnectCookieUnchecked;
}

/// Reply type for [Authenticate].
///
/// Can be obtained from a [AuthenticateCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [AuthenticateCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct AuthenticateReply {
    raw: *const u8,
}

impl AuthenticateReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn authenticated(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for AuthenticateReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for AuthenticateReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("AuthenticateReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("authenticated", &self.authenticated())
            .finish()
    }
}

impl Drop for AuthenticateReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for AuthenticateReply {}
unsafe impl std::marker::Sync for AuthenticateReply {}

/// Cookie type for [Authenticate].
///
/// This cookie can be used to get a [AuthenticateReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct AuthenticateCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [Authenticate].
///
/// This cookie can be used to get a [AuthenticateReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct AuthenticateCookieUnchecked {
    seq: u64,
}

impl base::Cookie for AuthenticateCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        AuthenticateCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for AuthenticateCookie {
}

unsafe impl base::CookieWithReplyChecked for AuthenticateCookie {
    type Reply = AuthenticateReply;
}

impl base::Cookie for AuthenticateCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        AuthenticateCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for AuthenticateCookieUnchecked {
    type Reply = AuthenticateReply;
}

/// The `Authenticate` request.
///
/// This request replies [AuthenticateReply].
///
/// Associated cookie types are [AuthenticateCookie] and [AuthenticateCookieUnchecked].
#[derive(Clone, Debug)]
pub struct Authenticate {
    pub window: xproto::Window,
    pub magic: u32,
}

unsafe impl base::RawRequest for Authenticate {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 2,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 12];
        self.window.serialize(&mut buf0[4 .. ]);
        self.magic.serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for Authenticate {
    type Cookie = AuthenticateCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for Authenticate {
    type Reply = AuthenticateReply;
    type Cookie = AuthenticateCookie;
    type CookieUnchecked = AuthenticateCookieUnchecked;
}

/// The `CreateDrawable` request.
///
/// This request has no reply.
///
/// Associated cookie types are [VoidCookie](crate::VoidCookie) and [VoidCookieChecked](crate::VoidCookieChecked).
#[derive(Clone, Debug)]
pub struct CreateDrawable {
    pub drawable: xproto::Drawable,
}

unsafe impl base::RawRequest for CreateDrawable {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 3,
            isvoid: 1,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 8];
        self.drawable.serialize(&mut buf0[4 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 8;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for CreateDrawable {
    type Cookie = base::VoidCookie;
    const IS_VOID: bool = true;
}

impl base::RequestWithoutReply for CreateDrawable {
}

/// The `DestroyDrawable` request.
///
/// This request has no reply.
///
/// Associated cookie types are [VoidCookie](crate::VoidCookie) and [VoidCookieChecked](crate::VoidCookieChecked).
#[derive(Clone, Debug)]
pub struct DestroyDrawable {
    pub drawable: xproto::Drawable,
}

unsafe impl base::RawRequest for DestroyDrawable {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 4,
            isvoid: 1,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 8];
        self.drawable.serialize(&mut buf0[4 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 8;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for DestroyDrawable {
    type Cookie = base::VoidCookie;
    const IS_VOID: bool = true;
}

impl base::RequestWithoutReply for DestroyDrawable {
}

/// Reply type for [GetBuffers].
///
/// Can be obtained from a [GetBuffersCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [GetBuffersCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetBuffersReply {
    raw: *const u8,
}

impl GetBuffersReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    unsafe fn compute_len(ptr: *const u8, _params: ()) -> usize {
        let mut sz = 0;
        // response_type
        sz += 1usize;
        // pad
        sz += 1usize;
        // sequence
        sz += 2usize;
        // length
        sz += 4usize;
        // width
        sz += 4usize;
        // height
        sz += 4usize;
        // count
        let count = *(ptr.add(sz) as *const u32);
        sz += 4usize;
        // pad
        sz += 12usize;
        // buffers
        sz += ((count as usize) * 20usize);
        sz
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn width(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn height(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    fn count(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }


    pub fn buffers(&self) -> &[Dri2Buffer] {
        unsafe {
            let offset = 32usize;
            let len = (self.count() as usize) as _;
            let ptr = self.wire_ptr().add(offset) as *const Dri2Buffer;
            std::slice::from_raw_parts(ptr, len)
        }
    }
}

impl base::Reply for GetBuffersReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for GetBuffersReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GetBuffersReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("width", &self.width())
            .field("height", &self.height())
            .field("count", &self.count())
            .field("pad", &12)
            .field("buffers", &self.buffers())
            .finish()
    }
}

impl Drop for GetBuffersReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for GetBuffersReply {}
unsafe impl std::marker::Sync for GetBuffersReply {}

/// Cookie type for [GetBuffers].
///
/// This cookie can be used to get a [GetBuffersReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct GetBuffersCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [GetBuffers].
///
/// This cookie can be used to get a [GetBuffersReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetBuffersCookieUnchecked {
    seq: u64,
}

impl base::Cookie for GetBuffersCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetBuffersCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for GetBuffersCookie {
}

unsafe impl base::CookieWithReplyChecked for GetBuffersCookie {
    type Reply = GetBuffersReply;
}

impl base::Cookie for GetBuffersCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetBuffersCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for GetBuffersCookieUnchecked {
    type Reply = GetBuffersReply;
}

/// The `GetBuffers` request.
///
/// This request replies [GetBuffersReply].
///
/// Associated cookie types are [GetBuffersCookie] and [GetBuffersCookieUnchecked].
#[derive(Clone, Debug)]
pub struct GetBuffers<'a> {
    pub drawable: xproto::Drawable,
    pub count: u32,
    pub attachments: &'a [u32],
}

unsafe impl<'a> base::RawRequest for GetBuffers<'a> {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 4,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 5,
            isvoid: 0,
        };

        let mut sections: [iovec; 6] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 6];

        let buf0: &mut [u8] = &mut [0; 12];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.count.serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        sections[4].iov_base = self.attachments.as_ptr() as *mut _;
        sections[4].iov_len = self.attachments.len() * std::mem::size_of::<u32>();
        sections[5].iov_len = base::align_pad(sections[4].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl<'a> base::Request for GetBuffers<'a> {
    type Cookie = GetBuffersCookie;
    const IS_VOID: bool = false;
}

impl<'a> base::RequestWithReply for GetBuffers<'a> {
    type Reply = GetBuffersReply;
    type Cookie = GetBuffersCookie;
    type CookieUnchecked = GetBuffersCookieUnchecked;
}

/// Reply type for [CopyRegion].
///
/// Can be obtained from a [CopyRegionCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [CopyRegionCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct CopyRegionReply {
    raw: *const u8,
}

impl CopyRegionReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for CopyRegionReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for CopyRegionReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("CopyRegionReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .finish()
    }
}

impl Drop for CopyRegionReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for CopyRegionReply {}
unsafe impl std::marker::Sync for CopyRegionReply {}

/// Cookie type for [CopyRegion].
///
/// This cookie can be used to get a [CopyRegionReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct CopyRegionCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [CopyRegion].
///
/// This cookie can be used to get a [CopyRegionReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct CopyRegionCookieUnchecked {
    seq: u64,
}

impl base::Cookie for CopyRegionCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        CopyRegionCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for CopyRegionCookie {
}

unsafe impl base::CookieWithReplyChecked for CopyRegionCookie {
    type Reply = CopyRegionReply;
}

impl base::Cookie for CopyRegionCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        CopyRegionCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for CopyRegionCookieUnchecked {
    type Reply = CopyRegionReply;
}

/// The `CopyRegion` request.
///
/// This request replies [CopyRegionReply].
///
/// Associated cookie types are [CopyRegionCookie] and [CopyRegionCookieUnchecked].
#[derive(Clone, Debug)]
pub struct CopyRegion {
    pub drawable: xproto::Drawable,
    pub region: u32,
    pub dest: u32,
    pub src: u32,
}

unsafe impl base::RawRequest for CopyRegion {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 6,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 20];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.region.serialize(&mut buf0[8 .. ]);
        self.dest.serialize(&mut buf0[12 .. ]);
        self.src.serialize(&mut buf0[16 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 20;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for CopyRegion {
    type Cookie = CopyRegionCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for CopyRegion {
    type Reply = CopyRegionReply;
    type Cookie = CopyRegionCookie;
    type CookieUnchecked = CopyRegionCookieUnchecked;
}

/// Reply type for [GetBuffersWithFormat].
///
/// Can be obtained from a [GetBuffersWithFormatCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [GetBuffersWithFormatCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetBuffersWithFormatReply {
    raw: *const u8,
}

impl GetBuffersWithFormatReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    unsafe fn compute_len(ptr: *const u8, _params: ()) -> usize {
        let mut sz = 0;
        // response_type
        sz += 1usize;
        // pad
        sz += 1usize;
        // sequence
        sz += 2usize;
        // length
        sz += 4usize;
        // width
        sz += 4usize;
        // height
        sz += 4usize;
        // count
        let count = *(ptr.add(sz) as *const u32);
        sz += 4usize;
        // pad
        sz += 12usize;
        // buffers
        sz += ((count as usize) * 20usize);
        sz
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn width(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn height(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    fn count(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }


    pub fn buffers(&self) -> &[Dri2Buffer] {
        unsafe {
            let offset = 32usize;
            let len = (self.count() as usize) as _;
            let ptr = self.wire_ptr().add(offset) as *const Dri2Buffer;
            std::slice::from_raw_parts(ptr, len)
        }
    }
}

impl base::Reply for GetBuffersWithFormatReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for GetBuffersWithFormatReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GetBuffersWithFormatReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("width", &self.width())
            .field("height", &self.height())
            .field("count", &self.count())
            .field("pad", &12)
            .field("buffers", &self.buffers())
            .finish()
    }
}

impl Drop for GetBuffersWithFormatReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for GetBuffersWithFormatReply {}
unsafe impl std::marker::Sync for GetBuffersWithFormatReply {}

/// Cookie type for [GetBuffersWithFormat].
///
/// This cookie can be used to get a [GetBuffersWithFormatReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct GetBuffersWithFormatCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [GetBuffersWithFormat].
///
/// This cookie can be used to get a [GetBuffersWithFormatReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetBuffersWithFormatCookieUnchecked {
    seq: u64,
}

impl base::Cookie for GetBuffersWithFormatCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetBuffersWithFormatCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for GetBuffersWithFormatCookie {
}

unsafe impl base::CookieWithReplyChecked for GetBuffersWithFormatCookie {
    type Reply = GetBuffersWithFormatReply;
}

impl base::Cookie for GetBuffersWithFormatCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetBuffersWithFormatCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for GetBuffersWithFormatCookieUnchecked {
    type Reply = GetBuffersWithFormatReply;
}

/// The `GetBuffersWithFormat` request.
///
/// This request replies [GetBuffersWithFormatReply].
///
/// Associated cookie types are [GetBuffersWithFormatCookie] and [GetBuffersWithFormatCookieUnchecked].
#[derive(Clone, Debug)]
pub struct GetBuffersWithFormat<'a> {
    pub drawable: xproto::Drawable,
    pub count: u32,
    pub attachments: &'a [AttachFormat],
}

unsafe impl<'a> base::RawRequest for GetBuffersWithFormat<'a> {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 4,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 7,
            isvoid: 0,
        };

        let mut sections: [iovec; 6] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 6];

        let buf0: &mut [u8] = &mut [0; 12];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.count.serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        sections[4].iov_base = self.attachments.as_ptr() as *mut _;
        sections[4].iov_len = self.attachments.len() * std::mem::size_of::<AttachFormat>();
        sections[5].iov_len = base::align_pad(sections[4].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl<'a> base::Request for GetBuffersWithFormat<'a> {
    type Cookie = GetBuffersWithFormatCookie;
    const IS_VOID: bool = false;
}

impl<'a> base::RequestWithReply for GetBuffersWithFormat<'a> {
    type Reply = GetBuffersWithFormatReply;
    type Cookie = GetBuffersWithFormatCookie;
    type CookieUnchecked = GetBuffersWithFormatCookieUnchecked;
}

/// Reply type for [SwapBuffers].
///
/// Can be obtained from a [SwapBuffersCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [SwapBuffersCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct SwapBuffersReply {
    raw: *const u8,
}

impl SwapBuffersReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn swap_hi(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn swap_lo(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for SwapBuffersReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for SwapBuffersReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("SwapBuffersReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("swap_hi", &self.swap_hi())
            .field("swap_lo", &self.swap_lo())
            .finish()
    }
}

impl Drop for SwapBuffersReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for SwapBuffersReply {}
unsafe impl std::marker::Sync for SwapBuffersReply {}

/// Cookie type for [SwapBuffers].
///
/// This cookie can be used to get a [SwapBuffersReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct SwapBuffersCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [SwapBuffers].
///
/// This cookie can be used to get a [SwapBuffersReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct SwapBuffersCookieUnchecked {
    seq: u64,
}

impl base::Cookie for SwapBuffersCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        SwapBuffersCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for SwapBuffersCookie {
}

unsafe impl base::CookieWithReplyChecked for SwapBuffersCookie {
    type Reply = SwapBuffersReply;
}

impl base::Cookie for SwapBuffersCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        SwapBuffersCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for SwapBuffersCookieUnchecked {
    type Reply = SwapBuffersReply;
}

/// The `SwapBuffers` request.
///
/// This request replies [SwapBuffersReply].
///
/// Associated cookie types are [SwapBuffersCookie] and [SwapBuffersCookieUnchecked].
#[derive(Clone, Debug)]
pub struct SwapBuffers {
    pub drawable: xproto::Drawable,
    pub target_msc_hi: u32,
    pub target_msc_lo: u32,
    pub divisor_hi: u32,
    pub divisor_lo: u32,
    pub remainder_hi: u32,
    pub remainder_lo: u32,
}

unsafe impl base::RawRequest for SwapBuffers {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 8,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 32];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.target_msc_hi.serialize(&mut buf0[8 .. ]);
        self.target_msc_lo.serialize(&mut buf0[12 .. ]);
        self.divisor_hi.serialize(&mut buf0[16 .. ]);
        self.divisor_lo.serialize(&mut buf0[20 .. ]);
        self.remainder_hi.serialize(&mut buf0[24 .. ]);
        self.remainder_lo.serialize(&mut buf0[28 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 32;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for SwapBuffers {
    type Cookie = SwapBuffersCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for SwapBuffers {
    type Reply = SwapBuffersReply;
    type Cookie = SwapBuffersCookie;
    type CookieUnchecked = SwapBuffersCookieUnchecked;
}

/// Reply type for [GetMsc].
///
/// Can be obtained from a [GetMscCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [GetMscCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetMscReply {
    raw: *const u8,
}

impl GetMscReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_hi(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_lo(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_hi(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_lo(&self) -> u32 {
        unsafe {
            let offset = 20usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc_hi(&self) -> u32 {
        unsafe {
            let offset = 24usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc_lo(&self) -> u32 {
        unsafe {
            let offset = 28usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for GetMscReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for GetMscReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GetMscReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("ust_hi", &self.ust_hi())
            .field("ust_lo", &self.ust_lo())
            .field("msc_hi", &self.msc_hi())
            .field("msc_lo", &self.msc_lo())
            .field("sbc_hi", &self.sbc_hi())
            .field("sbc_lo", &self.sbc_lo())
            .finish()
    }
}

impl Drop for GetMscReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for GetMscReply {}
unsafe impl std::marker::Sync for GetMscReply {}

/// Cookie type for [GetMsc].
///
/// This cookie can be used to get a [GetMscReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct GetMscCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [GetMsc].
///
/// This cookie can be used to get a [GetMscReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetMscCookieUnchecked {
    seq: u64,
}

impl base::Cookie for GetMscCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetMscCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for GetMscCookie {
}

unsafe impl base::CookieWithReplyChecked for GetMscCookie {
    type Reply = GetMscReply;
}

impl base::Cookie for GetMscCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetMscCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for GetMscCookieUnchecked {
    type Reply = GetMscReply;
}

/// The `GetMsc` request.
///
/// This request replies [GetMscReply].
///
/// Associated cookie types are [GetMscCookie] and [GetMscCookieUnchecked].
#[derive(Clone, Debug)]
pub struct GetMsc {
    pub drawable: xproto::Drawable,
}

unsafe impl base::RawRequest for GetMsc {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 9,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 8];
        self.drawable.serialize(&mut buf0[4 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 8;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for GetMsc {
    type Cookie = GetMscCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for GetMsc {
    type Reply = GetMscReply;
    type Cookie = GetMscCookie;
    type CookieUnchecked = GetMscCookieUnchecked;
}

/// Reply type for [WaitMsc].
///
/// Can be obtained from a [WaitMscCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [WaitMscCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct WaitMscReply {
    raw: *const u8,
}

impl WaitMscReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_hi(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_lo(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_hi(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_lo(&self) -> u32 {
        unsafe {
            let offset = 20usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc_hi(&self) -> u32 {
        unsafe {
            let offset = 24usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc_lo(&self) -> u32 {
        unsafe {
            let offset = 28usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for WaitMscReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for WaitMscReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("WaitMscReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("ust_hi", &self.ust_hi())
            .field("ust_lo", &self.ust_lo())
            .field("msc_hi", &self.msc_hi())
            .field("msc_lo", &self.msc_lo())
            .field("sbc_hi", &self.sbc_hi())
            .field("sbc_lo", &self.sbc_lo())
            .finish()
    }
}

impl Drop for WaitMscReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for WaitMscReply {}
unsafe impl std::marker::Sync for WaitMscReply {}

/// Cookie type for [WaitMsc].
///
/// This cookie can be used to get a [WaitMscReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct WaitMscCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [WaitMsc].
///
/// This cookie can be used to get a [WaitMscReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct WaitMscCookieUnchecked {
    seq: u64,
}

impl base::Cookie for WaitMscCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        WaitMscCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for WaitMscCookie {
}

unsafe impl base::CookieWithReplyChecked for WaitMscCookie {
    type Reply = WaitMscReply;
}

impl base::Cookie for WaitMscCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        WaitMscCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for WaitMscCookieUnchecked {
    type Reply = WaitMscReply;
}

/// The `WaitMsc` request.
///
/// This request replies [WaitMscReply].
///
/// Associated cookie types are [WaitMscCookie] and [WaitMscCookieUnchecked].
#[derive(Clone, Debug)]
pub struct WaitMsc {
    pub drawable: xproto::Drawable,
    pub target_msc_hi: u32,
    pub target_msc_lo: u32,
    pub divisor_hi: u32,
    pub divisor_lo: u32,
    pub remainder_hi: u32,
    pub remainder_lo: u32,
}

unsafe impl base::RawRequest for WaitMsc {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 10,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 32];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.target_msc_hi.serialize(&mut buf0[8 .. ]);
        self.target_msc_lo.serialize(&mut buf0[12 .. ]);
        self.divisor_hi.serialize(&mut buf0[16 .. ]);
        self.divisor_lo.serialize(&mut buf0[20 .. ]);
        self.remainder_hi.serialize(&mut buf0[24 .. ]);
        self.remainder_lo.serialize(&mut buf0[28 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 32;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for WaitMsc {
    type Cookie = WaitMscCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for WaitMsc {
    type Reply = WaitMscReply;
    type Cookie = WaitMscCookie;
    type CookieUnchecked = WaitMscCookieUnchecked;
}

/// Reply type for [WaitSbc].
///
/// Can be obtained from a [WaitSbcCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [WaitSbcCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct WaitSbcReply {
    raw: *const u8,
}

impl WaitSbcReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }


    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_hi(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn ust_lo(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_hi(&self) -> u32 {
        unsafe {
            let offset = 16usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn msc_lo(&self) -> u32 {
        unsafe {
            let offset = 20usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc_hi(&self) -> u32 {
        unsafe {
            let offset = 24usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn sbc_lo(&self) -> u32 {
        unsafe {
            let offset = 28usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for WaitSbcReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for WaitSbcReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("WaitSbcReply")
            .field("response_type", &self.response_type())
            .field("pad", &1)
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("ust_hi", &self.ust_hi())
            .field("ust_lo", &self.ust_lo())
            .field("msc_hi", &self.msc_hi())
            .field("msc_lo", &self.msc_lo())
            .field("sbc_hi", &self.sbc_hi())
            .field("sbc_lo", &self.sbc_lo())
            .finish()
    }
}

impl Drop for WaitSbcReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for WaitSbcReply {}
unsafe impl std::marker::Sync for WaitSbcReply {}

/// Cookie type for [WaitSbc].
///
/// This cookie can be used to get a [WaitSbcReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct WaitSbcCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [WaitSbc].
///
/// This cookie can be used to get a [WaitSbcReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct WaitSbcCookieUnchecked {
    seq: u64,
}

impl base::Cookie for WaitSbcCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        WaitSbcCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for WaitSbcCookie {
}

unsafe impl base::CookieWithReplyChecked for WaitSbcCookie {
    type Reply = WaitSbcReply;
}

impl base::Cookie for WaitSbcCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        WaitSbcCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for WaitSbcCookieUnchecked {
    type Reply = WaitSbcReply;
}

/// The `WaitSbc` request.
///
/// This request replies [WaitSbcReply].
///
/// Associated cookie types are [WaitSbcCookie] and [WaitSbcCookieUnchecked].
#[derive(Clone, Debug)]
pub struct WaitSbc {
    pub drawable: xproto::Drawable,
    pub target_sbc_hi: u32,
    pub target_sbc_lo: u32,
}

unsafe impl base::RawRequest for WaitSbc {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 11,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 16];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.target_sbc_hi.serialize(&mut buf0[8 .. ]);
        self.target_sbc_lo.serialize(&mut buf0[12 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 16;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for WaitSbc {
    type Cookie = WaitSbcCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for WaitSbc {
    type Reply = WaitSbcReply;
    type Cookie = WaitSbcCookie;
    type CookieUnchecked = WaitSbcCookieUnchecked;
}

/// The `SwapInterval` request.
///
/// This request has no reply.
///
/// Associated cookie types are [VoidCookie](crate::VoidCookie) and [VoidCookieChecked](crate::VoidCookieChecked).
#[derive(Clone, Debug)]
pub struct SwapInterval {
    pub drawable: xproto::Drawable,
    pub interval: u32,
}

unsafe impl base::RawRequest for SwapInterval {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 12,
            isvoid: 1,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 12];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.interval.serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for SwapInterval {
    type Cookie = base::VoidCookie;
    const IS_VOID: bool = true;
}

impl base::RequestWithoutReply for SwapInterval {
}

/// Reply type for [GetParam].
///
/// Can be obtained from a [GetParamCookie] with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
/// or from a [GetParamCookieUnchecked] with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetParamReply {
    raw: *const u8,
}

impl GetParamReply {

    fn wire_ptr(&self) -> *const u8 {
        self.raw
    }

    fn wire_len(&self) -> usize {
        (32 + self.length() * 4) as _
    }

    pub fn response_type(&self) -> u8 {
        unsafe {
            let offset = 0usize;
            let ptr = self.wire_ptr().add(offset) as *const u8;
            base::value_from_ptr(ptr)
        }
    }

    pub fn is_param_recognized(&self) -> bool {
        let val = unsafe { *(self.wire_ptr().add(1usize)) };
        val != 0
    }

    pub fn sequence(&self) -> u16 {
        unsafe {
            let offset = 2usize;
            let ptr = self.wire_ptr().add(offset) as *const u16;
            base::value_from_ptr(ptr)
        }
    }

    pub fn length(&self) -> u32 {
        unsafe {
            let offset = 4usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn value_hi(&self) -> u32 {
        unsafe {
            let offset = 8usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }

    pub fn value_lo(&self) -> u32 {
        unsafe {
            let offset = 12usize;
            let ptr = self.wire_ptr().add(offset) as *const u32;
            base::value_from_ptr(ptr)
        }
    }
}

impl base::Reply for GetParamReply {
    unsafe fn from_raw(raw: *const u8) -> Self {
        Self { raw }
    }

    unsafe fn into_raw(self) -> *const u8 {
        let raw = self.raw;
        std::mem::forget(self);
        raw
    }

    unsafe fn as_raw(&self) -> *const u8 {
        self.raw
    }
}

impl std::fmt::Debug for GetParamReply {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("GetParamReply")
            .field("response_type", &self.response_type())
            .field("is_param_recognized", &self.is_param_recognized())
            .field("sequence", &self.sequence())
            .field("length", &self.length())
            .field("value_hi", &self.value_hi())
            .field("value_lo", &self.value_lo())
            .finish()
    }
}

impl Drop for GetParamReply {
    fn drop(&mut self) {
        unsafe { libc::free(self.raw as *mut _); }
    }
}

unsafe impl std::marker::Send for GetParamReply {}
unsafe impl std::marker::Sync for GetParamReply {}

/// Cookie type for [GetParam].
///
/// This cookie can be used to get a [GetParamReply]
/// with [Connection::wait_for_reply](crate::Connection::wait_for_reply)
#[derive(Debug)]
pub struct GetParamCookie {
    seq: u64,
}

#[derive(Debug)]
/// Unchecked cookie type for [GetParam].
///
/// This cookie can be used to get a [GetParamReply]
/// with [Connection::wait_for_reply_unchecked](crate::Connection::wait_for_reply_unchecked)
pub struct GetParamCookieUnchecked {
    seq: u64,
}

impl base::Cookie for GetParamCookie {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetParamCookie { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieChecked for GetParamCookie {
}

unsafe impl base::CookieWithReplyChecked for GetParamCookie {
    type Reply = GetParamReply;
}

impl base::Cookie for GetParamCookieUnchecked {
    unsafe fn from_sequence(seq: u64) -> Self {
        GetParamCookieUnchecked { seq }
    }

    fn sequence(&self) -> u64 {
        self.seq
    }
}

unsafe impl base::CookieWithReplyUnchecked for GetParamCookieUnchecked {
    type Reply = GetParamReply;
}

/// The `GetParam` request.
///
/// This request replies [GetParamReply].
///
/// Associated cookie types are [GetParamCookie] and [GetParamCookieUnchecked].
#[derive(Clone, Debug)]
pub struct GetParam {
    pub drawable: xproto::Drawable,
    pub param: u32,
}

unsafe impl base::RawRequest for GetParam {
    fn raw_request(&self, c: &base::Connection, checked: bool) -> u64 { unsafe {
        let mut protocol_request = xcb_protocol_request_t {
            count: 2,
            ext: std::ptr::addr_of_mut!(FFI_EXT),
            opcode: 13,
            isvoid: 0,
        };

        let mut sections: [iovec; 4] = [iovec {
            iov_base: std::ptr::null_mut(),
            iov_len: 0,
        }; 4];

        let buf0: &mut [u8] = &mut [0; 12];
        self.drawable.serialize(&mut buf0[4 .. ]);
        self.param.serialize(&mut buf0[8 .. ]);
        sections[2].iov_base = buf0.as_mut_ptr() as *mut _;
        sections[2].iov_len = 12;
        sections[3].iov_len = base::align_pad(sections[2].iov_len, 4);

        let flags = if checked { base::RequestFlags::CHECKED } else { base::RequestFlags::NONE };

        xcb_send_request64(
            c.get_raw_conn(),
            flags.bits() as _,
            sections.as_mut_ptr().add(2),
            &mut protocol_request as *mut _,
        )
    }
}}

impl base::Request for GetParam {
    type Cookie = GetParamCookie;
    const IS_VOID: bool = false;
}

impl base::RequestWithReply for GetParam {
    type Reply = GetParamReply;
    type Cookie = GetParamCookie;
    type CookieUnchecked = GetParamCookieUnchecked;
}
